@@ -1,13 +1,13 @@
 #include <config.h>
 
-#include <apt-pkg/configuration.h>
 #include <apt-pkg/aptconfiguration.h>
+#include <apt-pkg/configuration.h>
 #include <apt-pkg/indexcopy.h>
 
+#include <cstdio>
 #include <string>
-#include <stdio.h>
 
-#include <gtest/gtest.h>
+#include "common.h"
 
 class NoCopy : private IndexCopy {
    public:
